﻿Option Strict On
Option Explicit On

Imports System.IO
Imports System.Runtime.InteropServices

Public Class TranscodeImage

#Region "API"
    <DllImport("Shlwapi.dll", EntryPoint:="SHCreateMemStream")>
    Private Shared Function SHCreateMemStream(<[In]> pInit As IntPtr,
                                              <[In]> cbInit As Integer) As <MarshalAs(UnmanagedType.Interface)> Object
    End Function

    <DllImport("Shlwapi.dll", EntryPoint:="SHCreateStreamOnFileEx")>
    <PreserveSig> Private Shared Function SHCreateStreamOnFileEx(<[In], MarshalAs(UnmanagedType.LPWStr)> pszFile As String,
                                                                 <[In]> grfMode As STGM,
                                                                 <[In]> dwAttributes As Integer,
                                                                 <[In], MarshalAs(UnmanagedType.Bool)> fCreate As Boolean,
                                                                 <[In]> pstmTemplate As IntPtr,
                                                                 <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppstm As Object) As Integer
    End Function

    <DllImport("Shell32.dll", EntryPoint:="SHCreateItemFromParsingName")>
    <PreserveSig> Private Shared Function SHCreateItemFromParsingName(<[In], MarshalAs(UnmanagedType.LPWStr)> pszPath As String,
                                                                      <[In]> pbc As IntPtr,
                                                                      <[In]> ByRef riid As Guid,
                                                                      <Out> ByRef pUnk As IntPtr) As Integer
    End Function

    <DllImport("Shell32.dll", EntryPoint:="SHGetPropertyStoreFromParsingName")>
    <PreserveSig> Private Shared Function SHGetPropertyStoreFromParsingName(<[In], MarshalAs(UnmanagedType.LPWStr)> pszPath As String,
                                                                            <[In]> pbc As IntPtr,
                                                                            <[In]> flags As GETPROPERTYSTOREFLAGS,
                                                                            <[In]> ByRef riid As Guid,
                                                                            <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppv As Object) As Integer
    End Function

    <DllImport("Propsys.dll", EntryPoint:="PSGetPropertyKeyFromName")>
    <PreserveSig> Private Shared Function PSGetPropertyKeyFromName(<[In], MarshalAs(UnmanagedType.LPWStr)> pszName As String,
                                                                   <Out> ByRef ppropkey As PROPERTYKEY) As Integer
    End Function

    <DllImport("Ole32.dll", EntryPoint:="PropVariantClear")>
    <PreserveSig> Private Shared Function PropVariantClear(<[In]> ByRef pVar As PROPVARIANT) As Integer
    End Function
#End Region

#Region "Enum"
    Public Enum TI_FLAGS As Integer
        TI_BITMAP = 1
        TI_JPEG = 2
    End Enum

    Private Enum STREAM_SEEK As Integer
        STREAM_SEEK_SET = 0
        STREAM_SEEK_CUR = 1
        STREAM_SEEK_END = 2
    End Enum

    Private Enum GETPROPERTYSTOREFLAGS As Integer
        GPS_DEFAULT = &H0
        GPS_HANDLERPROPERTIESONLY = &H1
        GPS_READWRITE = &H2
        GPS_TEMPORARY = &H4
        GPS_FASTPROPERTIESONLY = &H8
        GPS_OPENSLOWITEM = &H10
        GPS_DELAYCREATION = &H20
        GPS_BESTEFFORT = &H40
        GPS_NO_OPLOCK = &H80
        GPS_PREFERQUERYPROPERTIES = &H100
        GPS_EXTRINSICPROPERTIES = &H200
        GPS_EXTRINSICPROPERTIESONLY = &H400
        GPS_VOLATILEPROPERTIES = &H800
        GPS_VOLATILEPROPERTIESONLY = &H1000
        GPS_MASK_VALID = &H1FFF
    End Enum
    Private Enum STGM As Integer
        STGM_FAILIFTHERE = &H0
        STGM_DIRECT = &H0
        STGM_READ = &H0
        STGM_WRITE = &H1
        STGM_READWRITE = &H2
        STGM_SHARE_EXCLUSIVE = &H10
        STGM_SHARE_DENY_WRITE = &H20
        STGM_SHARE_DENY_READ = &H30
        STGM_SHARE_DENY_NONE = &H40
        STGM_CREATE = &H1000
        STGM_TRANSACTED = &H10000
        STGM_CONVERT = &H20000
        STGM_PRIORITY = &H40000
        STGM_NOSCRATCH = &H100000
        STGM_NOSNAPSHOT = &H200000
        STGM_DIRECT_SWMR = &H400000
        STGM_SIMPLE = &H8000000
        STGM_DELETEONRELEASE = &H4000000
    End Enum
#End Region

#Region "Const"
    Private Const S_OK As Integer = &H0
    Private Const VT_UI4 As Integer = &H13
    Private Const PropImageWidth As String = "System.Image.HorizontalSize"
    Private Const PropImageHeight As String = "System.Image.VerticalSize"

    Private Const IID_IStream As String = "0000000C-0000-0000-C000-000000000046"
    Private Const IID_IShellItem As String = "43826D1E-E718-42EE-BC55-A1E261C37BFE"
    Private Const IID_IPropertyStore As String = "886D8EEB-8CF2-4446-8D02-CDBA1DBDCF99"
    Private Const IID_ITranscodeImage As String = "BAE86DDD-DC11-421c-B7AB-CC55D1D65C44"
    Private Const CLSID_TranscodeImage As String = "17B75166-928F-417d-9685-64AA135565C1"
#End Region

#Region "Structure"
    <StructLayout(LayoutKind.Sequential)>
    Private Structure SizeU
        Dim Width As UInteger
        Dim Height As UInteger
        Sub New(Width As UInteger, Height As UInteger)
            Me.Width = Width
            Me.Height = Height
        End Sub
    End Structure

    <StructLayout(LayoutKind.Sequential)>
    Private Structure PROPERTYKEY
        Dim fmtid As Guid
        Dim pid As Integer
    End Structure

    <StructLayout(LayoutKind.Sequential)>
    Private Structure CArray
        Dim cElems As UInteger
        Dim pElems As IntPtr
    End Structure

    <StructLayout(LayoutKind.Explicit, Size:=16)>
    Private Structure PROPVARIANT
        <FieldOffset(0)> Dim vt As UShort
        <FieldOffset(2)> Dim wReserved1 As UShort
        <FieldOffset(4)> Dim wReserved2 As UShort
        <FieldOffset(6)> Dim wReserved3 As UShort
        <FieldOffset(8)> Dim ui4 As UInteger     'VT_UI4
        <FieldOffset(8)> Dim cArray As CArray    'VT_BLOB
    End Structure
#End Region

#Region "Public Functions"
    Public Function LoadImage(ImageFile As String, Optional NewImageWidthHeight As UInteger = 0) As Bitmap

        Dim bmpOut As Bitmap = Nothing

        If File.Exists(ImageFile) Then

            Dim bolOk As Boolean = False
            Dim ImageDimension As SizeU

            If NewImageWidthHeight = 0 Then

                ' Originaldimensionen des Bildes ermitteln
                If GetImageDimension(ImageFile, ImageDimension) Then bolOk = True

            Else

                ' neue Dimensionen setzen (Bild wird proportional skaliert)
                ImageDimension = New SizeU(NewImageWidthHeight, NewImageWidthHeight)
                bolOk = True

            End If

            If bolOk Then

                ' ein IShellItem vom Bild erstellen
                Dim pIShellItem As IntPtr = IntPtr.Zero
                If SHCreateItemFromParsingName(ImageFile, IntPtr.Zero, New Guid(IID_IShellItem), pIShellItem) = S_OK Then

                    If pIShellItem <> IntPtr.Zero Then

                        ' ein IStream (Memory) erstellen
                        Dim oIStream As Object = SHCreateMemStream(IntPtr.Zero, 0)

                        If oIStream IsNot Nothing Then

                            ' ein ITranscodeImage erstellen
                            Dim oITranscodeImage As Object = Activator.CreateInstance(Type.GetTypeFromCLSID(New Guid(CLSID_TranscodeImage)))

                            If oITranscodeImage IsNot Nothing Then

                                ' Bild transcodieren und in den IStream (Memory) schreiben
                                Dim uiRetWidth As UInteger
                                Dim uiRetHeight As UInteger
                                If CType(oITranscodeImage, ITranscodeImage).TranscodeImage(pIShellItem, ImageDimension.Width,
                                                                                           ImageDimension.Height, TI_FLAGS.TI_BITMAP,
                                                                                           oIStream, uiRetWidth, uiRetHeight) = S_OK Then

                                    ' Größe des Stream ermitteln
                                    Dim lngStreamSize As Long
                                    If CType(oIStream, IStream).Seek(0, STREAM_SEEK.STREAM_SEEK_END, lngStreamSize) = S_OK Then

                                        If lngStreamSize > 0 Then

                                            ' zurück zum Anfang des Streams
                                            If CType(oIStream, IStream).Seek(0, STREAM_SEEK.STREAM_SEEK_SET, 0) = S_OK Then

                                                ' ByteArray eentsprechend dimensionieren
                                                Dim uiReadBytes As UInteger
                                                Dim bytData As Byte() = New Byte(CInt(lngStreamSize) - 1) {}

                                                ' Daten von IStream in das ByteArray kopieren
                                                If CType(oIStream, IStream).Read(bytData, CUInt(lngStreamSize), uiReadBytes) = S_OK Then

                                                    ' alle Daten wurden kopiert?
                                                    If uiReadBytes = CUInt(lngStreamSize) Then

                                                        ' MemoryStream vom ByteArray erstellen
                                                        Using MemStream = New MemoryStream(bytData)

                                                            ' Bild vom MemoryStream erstellen
                                                            bmpOut = New Bitmap(MemStream)

                                                        End Using

                                                    End If

                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                                ' ITranscodeImage löschen
                                Marshal.ReleaseComObject(oITranscodeImage)

                            End If

                            ' IStream löschen
                            Marshal.ReleaseComObject(oIStream)

                        End If

                        ' IShellItem löschen
                        Marshal.Release(pIShellItem)

                    End If

                End If

            End If

        End If

        Return bmpOut

    End Function

    Public Function ConvertImage(ImageFile As String,
                                 Optional ConvertTo As TI_FLAGS = TI_FLAGS.TI_BITMAP,
                                 Optional NewImageWidthHeight As UInteger = 0) As Boolean

        Dim bolRet As Boolean = False

        If File.Exists(ImageFile) Then

            Dim bolOk As Boolean = False
            Dim ImageDimension As SizeU

            If NewImageWidthHeight = 0 Then

                If GetImageDimension(ImageFile, ImageDimension) Then bolOk = True

            Else

                ImageDimension = New SizeU(NewImageWidthHeight, NewImageWidthHeight)
                bolOk = True

            End If

            If bolOk Then

                ' ein IShellItem vom Bild erstellen
                Dim pIShellItem As IntPtr = IntPtr.Zero
                If SHCreateItemFromParsingName(ImageFile, IntPtr.Zero, New Guid(IID_IShellItem), pIShellItem) = S_OK Then

                    If pIShellItem <> IntPtr.Zero Then

                        ' Dateiname ersetzen
                        Select Case ConvertTo
                            Case TI_FLAGS.TI_JPEG
                                ImageFile = ImageFile.Replace(Path.GetFileName(ImageFile),
                                                          Path.GetFileNameWithoutExtension(ImageFile) &
                                                          "_converted_to" & ".jpg")
                            Case TI_FLAGS.TI_BITMAP
                                ImageFile = ImageFile.Replace(Path.GetFileName(ImageFile),
                                                          Path.GetFileNameWithoutExtension(ImageFile) &
                                                          "_converted_to" & ".bmp")
                        End Select

                        ' ein IStream (File) erstellen
                        Dim oIStream As Object = Nothing
                        If SHCreateStreamOnFileEx(ImageFile, STGM.STGM_CREATE Or STGM.STGM_WRITE, 0,
                                              False, IntPtr.Zero, oIStream) = S_OK Then

                            If oIStream IsNot Nothing Then

                                ' ein ITranscodeImage erstellen
                                Dim oITranscodeImage As Object = Activator.CreateInstance(Type.GetTypeFromCLSID(New Guid(CLSID_TranscodeImage)))

                                If oITranscodeImage IsNot Nothing Then

                                    ' Bild transcodieren und in den IStream (File) schreiben
                                    Dim uiRetWidth As UInteger
                                    Dim uiRetHeight As UInteger
                                    If CType(oITranscodeImage, ITranscodeImage).TranscodeImage(pIShellItem, ImageDimension.Width,
                                                                                               ImageDimension.Height, TI_FLAGS.TI_BITMAP,
                                                                                               oIStream, uiRetWidth, uiRetHeight) = S_OK Then

                                        bolRet = True

                                    End If

                                    ' ITranscodeImage löschen
                                    Marshal.ReleaseComObject(oITranscodeImage)

                                End If

                                ' IStream löschen
                                Marshal.ReleaseComObject(oIStream)

                            End If

                        End If

                        ' IShellItem löschen
                        Marshal.Release(pIShellItem)

                    End If

                End If

            End If

        End If

        Return bolRet

    End Function
#End Region

#Region "Private Functions"
    Private Function GetImageDimension(ImageFile As String, ByRef ImageDimension As SizeU) As Boolean

        Dim bolRet As Boolean = False

        ' PROPERTYKEY für System.Image.HorizontalSize ermitteln
        Dim tImageWidthKey As New PROPERTYKEY
        If PSGetPropertyKeyFromName(PropImageWidth, tImageWidthKey) = S_OK Then

            ' PROPERTYKEY für System.Image.VerticalSize ermitteln
            Dim tImageHeightKey As New PROPERTYKEY
            If PSGetPropertyKeyFromName(PropImageHeight, tImageHeightKey) = S_OK Then

                ' ein IPropertyStore vom Bild erstellen
                Dim oIPropertyStore As Object = Nothing
                If SHGetPropertyStoreFromParsingName(ImageFile, IntPtr.Zero, GETPROPERTYSTOREFLAGS.GPS_DEFAULT,
                                                     New Guid(IID_IPropertyStore), oIPropertyStore) = S_OK Then

                    If oIPropertyStore IsNot Nothing Then

                        ' PROPVARIANT vom PROPERTYKEY für System.Image.HorizontalSize ermitteln
                        Dim tImageWidthProp As New PROPVARIANT
                        If CType(oIPropertyStore, IPropertyStore).GetValue(tImageWidthKey, tImageWidthProp) = S_OK Then

                            ' PROPVARIANT vom PROPERTYKEY für System.Image.VerticalSize ermitteln
                            Dim tImageHeightProp As New PROPVARIANT
                            If CType(oIPropertyStore, IPropertyStore).GetValue(tImageHeightKey, tImageHeightProp) = S_OK Then

                                ' die PROPVARIANT.vt sind VT_UI4
                                If tImageWidthProp.vt = VT_UI4 Then
                                    If tImageHeightProp.vt = VT_UI4 Then

                                        ImageDimension = New SizeU(tImageWidthProp.ui4, tImageHeightProp.ui4)

                                        bolRet = True

                                    End If
                                End If

                                ' PROPVARIANT löschen
                                PropVariantClear(tImageHeightProp)

                            End If

                            ' PROPVARIANT löschen
                            PropVariantClear(tImageWidthProp)

                        End If

                        ' IPropertyStore löschen
                        Marshal.ReleaseComObject(oIPropertyStore)

                    End If

                End If

            End If

        End If

        Return bolRet

    End Function
#End Region

#Region "Interface IStream"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IStream)>
    Private Interface IStream

        <PreserveSig> Function Read(<Out, MarshalAs(UnmanagedType.LPArray, SizeParamIndex:=1)> pv As Byte(),
                                    <[In]> cb As UInteger,
                                    <Out> ByRef pcbRead As UInteger) As Integer

        <PreserveSig> Function Write() As Integer 'Dummy

        <PreserveSig> Function Seek(<[In]> dlibMove As Long,
                                    <[In]> dwOrigin As STREAM_SEEK,
                                    <Out> ByRef plibNewPosition As Long) As Integer

        'SetSize
        'CopyTo
        'Commit
        'Revert
        'LockRegion
        'UnlockRegion
        'Stat
        'Clone
    End Interface
#End Region

#Region "Interface IPropertyStore"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IPropertyStore)>
    Private Interface IPropertyStore
        <PreserveSig> Function GetCount() As Integer 'Dummy
        <PreserveSig> Function GetAt() As Integer 'Dummy
        <PreserveSig> Function GetValue(<[In]> ByRef key As PROPERTYKEY,
                                        <Out> ByRef pv As PROPVARIANT) As Integer
        'SetValue
        'Commit
    End Interface
#End Region

#Region "Interface ITranscodeImage"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_ITranscodeImage)>
    Private Interface ITranscodeImage
        <PreserveSig> Function TranscodeImage(<[In]> pShellItem As IntPtr,
                                              <[In]> uiMaxWidth As UInteger,
                                              <[In]> uiMaxHeight As UInteger,
                                              <[In]> flags As TI_FLAGS,
                                              <[In], MarshalAs(UnmanagedType.Interface)> pvImage As Object,
                                              <Out> ByRef puiWidth As UInteger,
                                              <Out> ByRef puiHeight As UInteger) As Integer
    End Interface
#End Region

End Class
